#include "twimMgr.h"


void
twimMgrCheckForWork(TWIM_MGR_INFO *infoPtr) {
  /* Check if job waiting in queue */
  volatile int zeroIfDataAvailable = k_fifo_is_empty(&(infoPtr->twimFifo));
//  printk("Check job queue: Fifo avail: %d\n",zeroIfDataAvailable);
  if( zeroIfDataAvailable == 0) {
//    printk("Getting TWIM job from queue\n");
    /* There is a job waiting; get it */
    TWIM_QUEUE_ITEM *newJob= k_fifo_get(&(infoPtr->twimFifo),K_FOREVER);
    /* Start it */
    twimStartJob(newJob);
  } else {
    /* No job waiting */
//    printk("No waiting TWIM job\n");
    infoPtr->currentJob= NULL;
  }
}

nrfx_twim_evt_t myEvent = {
  .type       = NRFX_TWIM_EVT_DONE
};

nrfx_err_t
twimStartJob(TWIM_QUEUE_ITEM *jobPtr) {
  nrfx_err_t err = NRFX_SUCCESS;
  // Set the busy indicator
  jobPtr->twimMgrInfoPtr->busy= 1;
  // Check if this job is a "normal" transfer or a special operation
  if(TWIM_SPECIAL_OPTION_CMD(jobPtr->flags) == 1) {
    // This is a special command - Determine type
    switch(TWIM_SPECIAL_OPTION_CMD(jobPtr->flags)) {
      // Drive the SCLK line low for the number of milliseconds specified
      case TWIM_SCLK_LOW: {
        // Disable TWIM so we can take control of the SCL pin
        nrfx_twim_disable(&(jobPtr->twimMgrInfoPtr->twimInst));
        // Configure the SCL pin as an output
        nrfx_gpiote_out_config_t sclPinConfig = NRFX_GPIOTE_CONFIG_OUT_SIMPLE(true);
        nrfx_gpiote_out_init(I2C_SCL_PIN,&sclPinConfig);
        // Drive the SCL pin low
        nrf_gpio_pin_clear(I2C_SCL_PIN);
        // Wait for the number of milliseconds as given in argument
        uint32_t delayMS= (uint32_t)(TWIM_SPECIAL_OPTION_VALUE(jobPtr->flags));
        k_msleep(delayMS);
        // Drive the SCL pin high
        nrf_gpio_pin_set(I2C_SCL_PIN);
        // Re-enable TWIM
        nrfx_twim_enable(&(jobPtr->twimMgrInfoPtr->twimInst));

        // Call the TWIM callback (since the TWIM wasn't really used here)
        // This callback will send an event to the Temp thread
        if(jobPtr->twim_cb != NULL) {
          jobPtr->twim_cb(&myEvent,jobPtr);
        }
      }
      break;
    }
    // Clear the busy indicator
    jobPtr->twimMgrInfoPtr->busy= 0;
      /* Check if job waiting in queue */
    twimMgrCheckForWork(jobPtr->twimMgrInfoPtr);
  } else {
    // This is a normal TWIM command
    // Start the job
    jobPtr->twimMgrInfoPtr->currentJob= jobPtr;
    err= nrfx_twim_xfer(&(jobPtr->twimMgrInfoPtr->twimInst),
                        &(jobPtr->xfer_desc),
                        TWIM_REGULAR_FLAGS(jobPtr->flags));
    // The TWIM will "run" in the background and the Handler will 
    // be called when it has completed; Busy indicator will be cleared there
    if(NRFX_SUCCESS != err) {
      printk("TWIM error during job start\n");
    }
  }
  //printk("       -- started\n");
  return err;
}

nrfx_err_t
twimScheduleJob(TWIM_QUEUE_ITEM *jobPtr) {
  nrfx_err_t err;
  //if(nrfx_twim_is_busy(&(jobPtr->twimMgrInfoPtr->twimInst)) == true) {
  if(jobPtr->twimMgrInfoPtr->busy == 1) {
    // TWIM is busy, so queue the job
//printk("Queueing TWIM job\n");
    k_fifo_put(&(jobPtr->twimMgrInfoPtr->twimFifo),jobPtr);
    err= NRFX_SUCCESS;
  } else {
    // Start the job
//printk("Starting TWIM job\n");
    err= twimStartJob(jobPtr);
  }
  return err;
}



#if 1

extern TWIM_MGR_INFO twimInfo_1;

/* Define the main handler for this TWIM */
void twimMgr_1_Handler(nrfx_twim_evt_t const *evtPtr,void *contextPtr) {
  /* TWIM has issued an event (completion); Deal with it */
  TWIM_MGR_INFO *infoPtr= &twimInfo_1;
  int *ptr= (int *)contextPtr;
//  printk("In handler: TWIM Event: %x  context: %x\n",evtPtr->type,*ptr);
  /* Check to see if there is a callback */
  TWIM_QUEUE_ITEM *currJob= infoPtr->currentJob;
  if(currJob->twim_cb != NULL) {
    /* There is a callback - call it */
    currJob->twim_cb(evtPtr,currJob);
  } else {
    /* No callback */
  }
  /* Clear the busy indicator */
  infoPtr->busy= 0;
  /* Check if job waiting in queue */
  twimMgrCheckForWork(infoPtr);
}

#endif