#ifndef _TWIMMGR_H_
#define _TWIMMGR_H_

#include "TeleMod.h"

struct twimQueueItemStruct;

typedef struct twimMgrInfoStruct {
  uint16_t                   index;               // The index used to create the TWIM instance
  uint16_t                   busy;                // Is the device busy
  nrfx_twim_t                twimInst;            // The instance structure for the created TWIM
  //nrfx_twim_t                *twimInstPtr;        // Pointer to TWIM
  nrfx_twim_config_t         twimConfig;          // Configuration information for the TWIM
  struct k_fifo              twimFifo;            // The FIFO for the TWIM job queue
  //struct k_fifo              *twimFifoPtr;        // Pointer to FIFO
  struct twimQueueItemStruct *currentJob;
} TWIM_MGR_INFO;

// "Job" structure for TWIM/I2C
typedef struct twimQueueItemStruct {
  void                       *fifo_reserved;      // For the FIFO
  struct twimMgrInfoStruct   *twimMgrInfoPtr;     // Pointer to TWIM instance info
  uint32_t                   flags;               // TWIM flags
  nrfx_twim_evt_handler_t    twim_cb;             // Callback function for this job
  uint32_t                   jobContext;          // Context (for this job)
  nrf_twim_frequency_t       speed;               // TWIM speed (for this job) 
  nrfx_twim_xfer_desc_t      xfer_desc;           // The TWIM xfer structure
} TWIM_QUEUE_ITEM;

#define TWIM_SPECIAL_OPTION_VALUE(x)  ((x & 0x00FF0000) >> 16)
#define TWIM_SPECIAL_OPTION_TYPE(x)   ((x & 0x7F000000) >> 24)
#define TWIM_SPECIAL_OPTION_CMD(x)    ((x & 0x80000000) >> 31)
#define TWIM_REGULAR_FLAGS(x)         ((x & 0x0000FFFF) >>  0)

#define TWIM_SCLK_LOW                 0x00000001

#define TWIM_SPECIAL_SCLK_LOW(lowMS)  ((0x80000000) | (TWIM_SCLK_LOW << 24) | (lowMS << 16))

#define TWIM_MGR_DEFINE(idx, sclPin, sdaPin)                                \
  /* Declare the TWIM info structure and populate it */                     \
  TWIM_MGR_INFO twimInfo_ ## idx = {                                        \
    .index                         = idx,                                   \
    .busy                          = 0,                                     \
    .twimInst                      = NRFX_TWIM_INSTANCE(idx),               \
    .twimConfig.scl                = sclPin,                                \
    .twimConfig.sda                = sdaPin,                                \
    .twimConfig.frequency          = NRF_TWIM_FREQ_400K,                    \
    .twimConfig.interrupt_priority = NRFX_TWIM_DEFAULT_CONFIG_IRQ_PRIORITY, \
    .twimConfig.hold_bus_uninit    = false,                                 \
    .currentJob                    = NULL                                   \
  };                                                                        \
                                                                            \
  /* Define a context item for this instance */                             \
  int twimMgr_ ## idx ## _ContextValue = 5;                                 \
  int *twimMgr_ ## idx ## _Context = &(twimMgr_ ## idx ## _ContextValue);   \
                                                                            \
                                                                            
#if 0
  /* Define the main handler for this TWIM */                               \
  void twimMgr_ ## idx ## _Handler(nrfx_twim_evt_t const *evtPtr,           \
                                   void *contextPtr) {                      \
    /* TWIM has issued an event (completion); Deal with it */               \
    TWIM_MGR_INFO *infoPtr= &(twimInfo_ ## idx);                            \
    int *ptr= (int *)contextPtr;                                            \
    printk("In handler: TWIM Event: %x  context: %x\n",evtPtr->type,*ptr);  \
    /* Check to see if there is a callback */                               \
    TWIM_QUEUE_ITEM *currJob= infoPtr->currentJob;                          \
    if(currJob->twim_cb != NULL) {                                          \
      /* There is a callback - call it */                                   \
      printf("Calling job callback\n");                                     \
      currJob->twim_cb(evtPtr,currJob);                                     \
    } else {                                                                \
      /* No callback */                                                     \
      printk("No TWIM callback\n");                                         \
    }                                                                       \
    /* Clear the busy indicator */                                          \
    infoPtr->busy= 0;                                                       \
    /* Check if job waiting in queue */                                     \
    twimMgrCheckForWork(infoPtr);                                           \
  }
#endif








#define TWIM_MGR_INIT(idx)                                                                  \
  /* extern nrfx_twim_evt_handler_t twimMgr_1_Handler;   */                                 \
  TWIM_MGR_INFO *infoPtr= &(twimInfo_ ## idx);                                              \
  /* Initialize the FIFO queue */                                                           \
  printk("Init FIFO\n");                                                                    \
  k_fifo_init(&(infoPtr->twimFifo));                                                        \
  /* Connect the interrupt to the system */                                                 \
  printk("Connect interrupt\n");                                                            \
  IRQ_CONNECT( SPIM ## idx ## _SPIS ## idx ## _TWIM ## idx ## _TWIS ## idx ## _SPI ## idx ## _TWI ## idx ## _IRQn , 6, nrfx_isr, nrfx_twim_ ## idx ## _irq_handler , 0 );                     \
  if(NRFX_SUCCESS != nrfx_twim_init(&(infoPtr->twimInst),                                   \
                                    &(infoPtr->twimConfig),                                 \
                                    (nrfx_twim_evt_handler_t) &twimMgr_ ## idx ##_Handler,  \
                                    (void *) &twimMgr_ ## idx ## _Context )     ) {         \
    printk("I2C driver init failed for index %x!\n",idx);                                   \
  }                                                                                         \
                                                                                            \
  /* Enable the interface */                                                                \
  nrfx_twim_enable(&(infoPtr->twimInst));


nrfx_err_t twimScheduleJob(TWIM_QUEUE_ITEM *jobPtr);
nrfx_err_t twimStartJob(TWIM_QUEUE_ITEM *jobPtr);





#endif


