/* main.c - Application main entry point */

/*
 * Copyright (c) 2015-2016 Intel Corporation
 *
 * SPDX-License-Identifier: Apache-2.0
 */

#include "TeleMod.h"

LOG_MODULE_REGISTER(TeleMod, CONFIG_LOG_DEFAULT_LEVEL);

struct bt_conn *bleCurrentConnection;

struct k_fifo notifyQueue;

static const struct bt_data ad[] = {
	BT_DATA_BYTES(BT_DATA_FLAGS, (BT_LE_AD_GENERAL | BT_LE_AD_NO_BREDR)),
	BT_DATA_BYTES(BT_DATA_UUID16_ALL,
		      BT_UUID_16_ENCODE(BT_UUID_HRS_VAL),
		      BT_UUID_16_ENCODE(BT_UUID_BAS_VAL),
		      BT_UUID_16_ENCODE(BT_UUID_DIS_VAL))
};

#if 0
// TWI (with transaction manager) initialization
static void twi_config(void) {
  uint32_t err_code;
  nrf_drv_twi_config_t const config = {
    .scl                  = NRF_GPIO_PIN_MAP(0,27),
    .sda                  = NRF_GPIO_PIN_MAP(0,26),
    .frequency            = NRF_DRV_TWI_FREQ_100K,
    .interrupt_priority   = APP_IRQ_PRIORITY_LOWEST,
    .clear_bus_init       = false
  };

  err_code = nrf_twi_mngr_init(&I2C0, &config);
  APP_ERROR_CHECK(err_code);
}
#endif

static struct bt_gatt_exchange_params exchange_params;

static void exchange_func(struct bt_conn *conn, uint8_t att_err,
			  struct bt_gatt_exchange_params *params) {
  printk("MTU exchange %s\n", att_err == 0 ? "successful" : "failed");
  printk("MTU size after mtu exchange: %d\n",bt_gatt_get_mtu(conn));
}

static void bleConnected(struct bt_conn *conn, uint8_t err) {
  static const struct bt_conn_le_data_len_param bigMTU = {
    .tx_max_len  = 247,
    .tx_max_time = 6500
  };
  //BT_CONN_LE_DATA_LEN_PARAM(247,20);
  if(err) {
    printk("Connection failed (err 0x%02x)\n", err);
  } else {
    bleCurrentConnection = bt_conn_ref(conn);
    // Try to change connection parameters
#if 1
    int sizeErr= bt_conn_le_data_len_update(conn,&bigMTU);
    if(sizeErr) {
      printk("MTU size change failed\n");
    }
#endif

    printk("Connected\n");
    printk("MTU size after len update: %d\n",bt_gatt_get_mtu(conn));

    exchange_params.func = exchange_func;

    int gattErr = bt_gatt_exchange_mtu(conn, &exchange_params);
    if (gattErr) {
      printk("MTU exchange failed (err %d)\n", gattErr);
    } else {
      printk("MTU exchange pending\n");
    }
  }
}

static void bleDisconnected(struct bt_conn *conn, uint8_t reason) {
	printk("Disconnected (reason 0x%02x)\n", reason);

	if (bleCurrentConnection) {
		bt_conn_unref(bleCurrentConnection);
		bleCurrentConnection = NULL;
	}
}

#ifdef CONFIG_BT_GATT_LBS_SECURITY_ENABLED
static void security_changed(struct bt_conn *conn, bt_security_t level,
			     enum bt_security_err err)
{
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	if (!err) {
		printk("Security changed: %s level %u\n", addr, level);
	} else {
		printk("Security failed: %s level %u err %d\n", addr, level,
			err);
	}
}
#endif

static struct bt_conn_cb bleConnectionCallbacks = {
	.connected    = bleConnected,
	.disconnected = bleDisconnected,
#ifdef CONFIG_BT_GATT_LBS_SECURITY_ENABLED
        .security_changed = security_changed,
#endif
};


#if defined(CONFIG_BT_GATT_LBS_SECURITY_ENABLED)
static void auth_passkey_display(struct bt_conn *conn, unsigned int passkey)
{
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	printk("Passkey for %s: %06u\n", addr, passkey);
}

static void auth_cancel(struct bt_conn *conn)
{
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	printk("Pairing cancelled: %s\n", addr);
}

static void pairing_confirm(struct bt_conn *conn)
{
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	bt_conn_auth_pairing_confirm(conn);

	printk("Pairing confirmed: %s\n", addr);
}

static void pairing_complete(struct bt_conn *conn, bool bonded)
{
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	printk("Pairing completed: %s, bonded: %d\n", addr, bonded);
}

static void pairing_failed(struct bt_conn *conn, enum bt_security_err reason)
{
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	printk("Pairing failed conn: %s, reason %d\n", addr, reason);
}
#endif



static void auth_cancel(struct bt_conn *conn) {
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	printk("Pairing cancelled: %s\n", addr);
}

static struct bt_conn_auth_cb bleConnectionAuthCallbacks = {
#if defined(CONFIG_BT_GATT_LBS_SECURITY_ENABLED)
	.passkey_display = auth_passkey_display,

	.pairing_confirm = pairing_confirm,
	.pairing_complete = pairing_complete,
	.pairing_failed = pairing_failed,
#endif
        .cancel = auth_cancel,
};


static void bleCreateServices(void) {
}


static void btAdvertisingStart(void) {
	int err;

	printk("Bluetooth initialized\n");

	err = bt_le_adv_start(BT_LE_ADV_CONN_NAME, ad, ARRAY_SIZE(ad), NULL, 0);
	if(err) {
		printk("Advertising failed to start (err %d)\n", err);
		return;
	}

	printk("Advertising successfully started\n");
}



#if 0
static void bas_notify(void)
{
	uint8_t battery_level = bt_bas_get_battery_level();

	battery_level--;

	if (!battery_level) {
		battery_level = 100U;
	}

	bt_bas_set_battery_level(battery_level);
}

static void hrs_notify(void)
{
	static uint8_t heartrate = 90U;

	/* Heartrate measurements simulation */
	heartrate++;
	if (heartrate == 160U) {
		heartrate = 90U;
	}

	bt_hrs_notify(heartrate);
}
#endif

#define MY_STACK_SIZE 1024
#define MIC_THREAD_PRIORITY      7
#define TEMP_THREAD_PRIORITY     7
#define BLE_SEND_THREAD_PRIORITY 7

extern void threadMicrophone(void *, void *, void *);
extern void threadThermometer(void *, void *, void *);
extern void threadBleSend(void *, void *, void *);

K_THREAD_STACK_DEFINE(threadMicrophone_stack, MY_STACK_SIZE);
K_THREAD_STACK_DEFINE(threadThermometer_stack, MY_STACK_SIZE);
K_THREAD_STACK_DEFINE(threadBleSend_stack, MY_STACK_SIZE);

struct k_thread threadMicrophone_tcb;
struct k_thread threadThermometer_tcb;
struct k_thread threadBleSend_tcb;

uint32_t bleNotificationsAvailable;

// Use twim directly with no management
#if 0
void
i2cHandler(nrfx_twim_evt_t const *evtPtr, void *contextPtr) {
  int *ptr= (int *)contextPtr;
  printk("TWIM Event: %x   context: %x\n",evtPtr->type,*ptr);
#if 1
  if(evtPtr->type == NRFX_TWIM_EVT_DONE) {
    printk("TWIM transfer finished\n");
    switch(*ptr) {
      case 2: {
#if 1
        int16_t tempCRaw = ((readBuf[0] << 8) | readBuf[1]) / 32;
        float tempC = (float)tempCRaw * 0.125;
        float tempF = (tempC * 1.8) + 32;
        char strBuf[30];
        sprintf(strBuf,"Temp F: %f   Temp C: %f\n",tempF,tempC);
        printk("%s\n",strBuf);
#endif
      }
      break;
    }
  }
#endif
}



  // I2C instance
nrfx_twim_t i2c1= NRFX_TWIM_INSTANCE(1);
int twimContext= 0;

extern int micContext;

#endif




#if 0
TWIM_MGR_DEFINE( 1, I2C_SCL_PIN, I2C_SDA_PIN );
#endif





#if 1
// Inline version of above DEFINE
/* Declare the TWIM info structure and populate it */
TWIM_MGR_INFO twimInfo_1 = {
  .index                         = 1,
  .busy                          = 0,
  .twimInst                      = NRFX_TWIM_INSTANCE(1),
  .twimConfig.scl                = I2C_SCL_PIN,
  .twimConfig.sda                = I2C_SDA_PIN,
  .twimConfig.frequency          = NRF_TWIM_FREQ_400K,
  .twimConfig.interrupt_priority = NRFX_TWIM_DEFAULT_CONFIG_IRQ_PRIORITY,
  .twimConfig.hold_bus_uninit    = false,
  .currentJob                    = NULL
};

  /* Define a context item for this instance */
  int twimMgr_1_Context = 5;    
                

#if 0

/* Define the main handler for this TWIM */
void twimMgr_1_Handler(nrfx_twim_evt_t const *evtPtr,void *contextPtr) {
  /* TWIM has issued an event (completion); Deal with it */
  TWIM_MGR_INFO *infoPtr= &twimInfo_1;
  int *ptr= (int *)contextPtr;
  printk("TWIM Event: %x  context: %x\n",evtPtr->type,*ptr);
  /* Check to see if there is a callback */
  TWIM_QUEUE_ITEM *currJob= infoPtr->currentJob;
  if(currJob->twim_cb != NULL) {
    /* There is a callback - call it */
    printf("Calling job callback\n");
    currJob->twim_cb(evtPtr,currJob);
  } else {
    /* No callback */
    printk("No TWIM callback\n");
  }
  /* Check if job waiting in queue */
  volatile int zeroIfDataAvailable = k_fifo_is_empty(infoPtr->twimFifoPtr);
  //volatile int zeroIfDataAvailable = k_fifo_is_empty(&(infoPtr->twimFifo));
  printk("Fifo avail: %d\n",zeroIfDataAvailable);
#if 0
  if( zeroIfDataAvailable == 0) {
    printk("Getting TWIM job from queue\n");
    /* There is a job waiting; get it */
    TWIM_QUEUE_ITEM *newJob= k_fifo_get(infoPtr->twimFifoPtr,K_FOREVER);
    /* Start it */
    twimStartJob(newJob);
  } else {
    /* No job waiting */
    printk("No waiting TWIM job\n");
#endif
    infoPtr->currentJob= NULL;
#if 0
  }
#endif
}

#endif

#if 1
static FATFS fat_fs;

/* mounting info */
static struct fs_mount_t mp = {
	.type = FS_FATFS,
	.fs_data = &fat_fs,
};

/*
*  Note the fatfs library is able to mount only strings inside _VOLUME_STRS
*  in ffconf.h
*/
static const char *disk_mount_pt = "/SD:";

#if 0
static struct fs_mount_t fs_mnt;

static int
setupFlash(struct fs_mount_t *mnt) {
  int returnCode = 0;

  unsigned int id;
  const struct flashArea *pfa;

  mnt->storage_dev= (void *)FLASH_AREA_ID(SDCARD);
  id= (uintptr_t)mnt->storage_dev;

  returnCode= flash_area_open(id, &fpa);
}

#endif
static int lsdir(const char *path)
{
	int res;
	struct fs_dir_t dirp;
	static struct fs_dirent entry;

	/* Verify fs_opendir() */
	res = fs_opendir(&dirp, path);
	if (res) {
		printk("Error opening dir %s [%d]\n", path, res);
		return res;
	}

	printk("\nListing dir %s ...\n", path);
	for (;;) {
		/* Verify fs_readdir() */
		res = fs_readdir(&dirp, &entry);

		/* entry.name[0] == 0 means end-of-dir */
		if (res || entry.name[0] == 0) {
			break;
		}

		if (entry.type == FS_DIR_ENTRY_DIR) {
			printk("[DIR ] %s\n", entry.name);
		} else {
			printk("[FILE] %s (size = %zu)\n",
				entry.name, entry.size);
		}
	}

	/* Verify fs_closedir() */
	fs_closedir(&dirp);

	return res;
}



static void
initDisk(void) {

#if 0
	/* raw disk i/o */
	do {
		static const char *disk_pdrv = "SD";
		uint64_t memory_size_mb;
		uint32_t block_count;
		uint32_t block_size;

		if (disk_access_init(disk_pdrv) != 0) {
			LOG_ERR("Storage init ERROR!");
			break;
		}

		if (disk_access_ioctl(disk_pdrv,
				DISK_IOCTL_GET_SECTOR_COUNT, &block_count)) {
			LOG_ERR("Unable to get sector count");
			break;
		}
		LOG_INF("Block count %u", block_count);

		if (disk_access_ioctl(disk_pdrv,
				DISK_IOCTL_GET_SECTOR_SIZE, &block_size)) {
			LOG_ERR("Unable to get sector size");
			break;
		}
		printk("Sector size %u\n", block_size);

		memory_size_mb = (uint64_t)block_count * block_size;
		printk("Memory Size(MB) %u\n", (uint32_t)(memory_size_mb >> 20));
	} while (0);
#endif
//printk("Done with ioctl\n");
	mp.mnt_point = disk_mount_pt;
printk("Trying to mount\n");
	int res = fs_mount(&mp);
printk("Mount done\n");
	if (res == FR_OK) {
		printk("Disk mounted.\n");
//		lsdir(disk_mount_pt);
	} else {
		printk("Error mounting disk.\n");
	}
#if 0
  struct fs_mount_t *mp = &fs_mnt;
  struct fs_dir_t   dir = { 0 };
  struct fs_statvfs sbuf;
  int returnCode;

  returnCode= setupFlash(mp);
#endif
}

#endif


// to here
#endif

//TWIM_MGR_INFO *iPtr= &twimInfo_1;

void main(void) {
  int err;

  // Initialize interfaces
  // GPIO
  if(!nrfx_gpiote_is_init()) {
    err= nrfx_gpiote_init(GPIO_INTERRUPT_LEVEL);
  }
  // Initialize Filesystem
  initDisk();


  lsdir(disk_mount_pt);

// Use twim directly with no management
#if 0
  // I2C configure
  nrfx_twim_config_t i2cConfig = {
    .scl =                 I2C_SCL_PIN,
    .sda =                 I2C_SDA_PIN,
    .frequency =           NRF_TWIM_FREQ_400K,
    .interrupt_priority =  NRFX_TWIM_DEFAULT_CONFIG_IRQ_PRIORITY,
    .hold_bus_uninit =     false
  };
  IRQ_CONNECT(SPIM1_SPIS1_TWIM1_TWIS1_SPI1_TWI1_IRQn,6,nrfx_isr,nrfx_twim_1_irq_handler,0);
  err= nrfx_twim_init(&i2c1,
                      &i2cConfig,
                      (nrfx_twim_evt_handler_t)i2cHandler,
                      //NULL,
                      (void *)&micContext);
  if(NRFX_SUCCESS != err) {
    printk("I2C driver init failed: %d\n", err);
  }
  nrfx_twim_enable(&i2c1);
#endif

  //

#if 0
  TWIM_MGR_INIT(1);
#endif

#if 1
// Inline version of above INIT for management version
  //extern nrfx_twim_evt_handler_t twimMgr_1_Handler(nrfx_twim_evt_t const *evtPtr, void *cPtr);
  extern nrfx_twim_evt_handler_t twimMgr_1_Handler;
  TWIM_MGR_INFO *infoPtr= &(twimInfo_1);
  printk("Init FIFO\n");
  /* Initialize the FIFO queue */
  k_fifo_init(&(infoPtr->twimFifo));
  /* Connect the interrupt to the system */
  printk("Connect interrupt\n");
  // For nRF52840
//  IRQ_CONNECT(SPIM1_SPIS1_TWIM1_TWIS1_SPI1_TWI1_IRQn,6,nrfx_isr,nrfx_twim_1_irq_handler,0);
  // For nRF5340
  IRQ_CONNECT(SPIM1_SPIS1_TWIM1_TWIS1_UARTE1_IRQn,6,nrfx_isr,nrfx_twim_1_irq_handler,0);
  //IRQ_CONNECT( SPIM ## idx ## _SPIS ## idx ## _TWIM ## idx ## _TWIS ## idx ## _SPI ## idx ## _TWI ## idx ## _IRQn , 6, nrfx_isr, nrfx_twim_ ## idx ## _irq_handler , 0 );
  if(NRFX_SUCCESS != nrfx_twim_init(&(infoPtr->twimInst),
                                    &(infoPtr->twimConfig),
                                    (nrfx_twim_evt_handler_t) &twimMgr_1_Handler,
                                    (void *) &twimMgr_1_Context )    ) {
    printk("I2C driver init failed!\n");
  }
  /* Enable the interface */
  nrfx_twim_enable(&(infoPtr->twimInst));
// to here
#endif

	err = bt_enable(NULL);
	if (err) {
		printk("Bluetooth init failed (err %d)\n", err);
		return;
	}
	// Create the services used by this app
        bleCreateServices();

	// Set up some callbacks
	bt_conn_cb_register(&bleConnectionCallbacks);
	bt_conn_auth_cb_register(&bleConnectionAuthCallbacks);

	// Start BLE advertising
	btAdvertisingStart();

        // Initialize modules/structures
        k_fifo_init(&notifyQueue);

        // Initialize variables
        bleNotificationsAvailable= 8;      // Must match  TX_BUF_COUNT in prj.conf

        // Define Threads/Tasks
// Use NULL,NULL,NULL  for standalone twim (no managment version)
// Use &twimInfo_1,NULL,NULL for management version to pass in &Info structure
        k_tid_t micTid = k_thread_create(&threadMicrophone_tcb, threadMicrophone_stack,
                                 K_THREAD_STACK_SIZEOF(threadMicrophone_stack),
                                 threadMicrophone,
                                 (void *)(&twimInfo_1), NULL, NULL,
                                 //NULL, NULL, NULL,
                                 MIC_THREAD_PRIORITY, 0, K_NO_WAIT);
        err= k_thread_name_set(micTid,"Mic");
        printk("ThreadNum: %d err: %d Name: %s\n",(int)micTid,err,k_thread_name_get(micTid));

#if 1
        k_tid_t tempTid = k_thread_create(&threadThermometer_tcb, threadThermometer_stack,
                                 K_THREAD_STACK_SIZEOF(threadThermometer_stack),
                                 threadThermometer,
                                 (void *)(&twimInfo_1), NULL, NULL,
                                 //NULL, NULL, NULL,
                                 TEMP_THREAD_PRIORITY, 0, K_NO_WAIT);
        err= k_thread_name_set(tempTid,"Temp");
        printk("ThreadNum: %d err: %d Name: %s\n",(int)tempTid,err,k_thread_name_get(tempTid));
#endif

        k_tid_t bleSendTid = k_thread_create(&threadBleSend_tcb, threadBleSend_stack,
                                 K_THREAD_STACK_SIZEOF(threadBleSend_stack),
                                 threadBleSend,
                                 NULL, NULL, NULL,
                                 BLE_SEND_THREAD_PRIORITY, 0, K_NO_WAIT);
        err= k_thread_name_set(bleSendTid,"BLEsend");
        printk("ThreadNum: %d err: %d Name: %s\n",(int)bleSendTid,err,k_thread_name_get(bleSendTid));


	/* Implement notification. At the moment there is no suitable way
	 * of starting delayed work so we do it here
	 */
	while (1) {
		k_sleep(K_SECONDS(1));
                k_thread_suspend(k_current_get());
#if 0
		/* Heartrate measurements simulation */
		hrs_notify();

		/* Battery level simulation */
		bas_notify();
#endif
	}
}
